
import express from "express";
import { Application, Request, Response, json as jsonBodyParser } from "express";
import morgan from 'morgan';

import { Product, ProductModel } from "./db/models";
import { QueryOperators, connect as connectDB } from "./db/connection";
import * as helpers from "./route.helpers";


/**
 * Gets all products in the database
 *
 * This endpoint accepts two parameters to filter the output:
 * - `priceFrom`: Define a minimum price of all products in the search
 * - `priceTo`: Define a maximum price of all products in the search
 *
 * ~~Responses~~
 * `200 OK` on success
 * @param req - request object
 * @param res - response object
 */
async function getProductsEndpoint(req: Request, res: Response): Promise<void> {
    const priceFromStr = req.query.priceFrom;
    const priceToStr = req.query.priceTo;
    let priceFrom: number = -1;
    let priceTo: number = -1;
    try {
        if (typeof priceFromStr === "string") {
            priceFrom = helpers.parseParamInt(priceFromStr);
        }
        if (typeof priceToStr === "string") {
            priceTo = helpers.parseParamInt(priceToStr);
        }
    } catch(error) {
        // Error processing numbers
        return res.status(400).end("One or more parameters could not be processed.");
    }

    // If these variables are below 0, then they are still the original number set before parsing
    const priceFromValid = priceFrom >= 0;
    const priceToValid = priceTo >= 0;

    if (priceFromValid && priceToValid) {
        if (priceFrom > priceTo) {
            return res.status(400).end("priceFrom parameter is bigger than priceTo parameter.");
        }
    }

    let options = {};
    if (priceFromValid || priceToValid) {
        // If at least one parameter is parseable
        const queryOperators: QueryOperators = {};
        if (priceFromValid) {
            queryOperators.$gte = priceFrom;
        }
        if (priceToValid) {
            queryOperators.$lte = priceTo;
        }
        options = {price: queryOperators};
    }

    const products = await ProductModel.find(options) as Product[];
    // Filter product objects in query to remove non-useful data
    const filteredProducts = products.map(product => { return Product.fromJson(product); });
    res.json(filteredProducts);
}

/**
 * Gets one product in the database using the unique identifier
 *
 * ~~Responses~~
 * `200 OK` on success
 * `404 Not Found` if product with identifier doesn't exist.
 * @param req - request object
 * @param res - response object
 */
async function getProductEndpoint(req: Request, res: Response): Promise<void> {
    const product = await ProductModel.findOne({identifier: req.params.identifier});
    if (!product) {
        return res.status(404).end(`Product with identifier ${req.params.identifier} could not be found.`);
    }
    // Run DB output through validator to strip any database specific keys (ObjectID's and version ID's)
    const validProduct: Product = Product.fromJson(product);
    res.json(validProduct);
}

/**
 * Creates a product. req.body is processed from middleware ran before this function. Parsed from raw data to a json object.
 *
 * ~~Responses~~
 * `200 OK` on success
 * `400 Bad Request` when trying to create an object with the same identifier as an existing object.
 * `401 Unauthorized` on request without header `X-Token` being set to an authorized key. See checkAuthorization middleware function.
 * @param req - request object
 * @param res - response object
 */
async function createProductEndpoint(req: Request, res: Response): Promise<void> {
    let postData: Product;
    try {
        postData = Product.fromJson(req.body);
    } catch (error) {
        // Thrown error if data received is not valid and cannot be inserted into DB
        return res.status(400).end("Request has invalid data. Cannot be added to database.");
    }

    const identifierAlreadyExists = await ProductModel.findOne({
        identifier: postData.identifier
    });
    if (identifierAlreadyExists) {
        return res.status(400).end("Supplied identifier already exists in database of products.");
    }
    try {
        ProductModel.create(postData);
        return res.status(200).end("Successfully added product to database.");
    } catch (error) {
        // Generic error handler for unexpected issue
        return res.status(500).end("Something unexpected happened and the server was unable to process your request.");
    }
}

/**
 * Edits an existing product, either partially or entirely based of user input
 *
 * ~~Responses~~
 * `200 OK` on success.
 * `400 Bad Request` if unable to parse some or all of the data received for updating.
 * `401 Unauthorized` on request without header `X-Token` being set to an authorized key.
 * `404 Not Found` if product with identifier doesn't exist.
 * @param req - request object
 * @param res - response object
 */
async function editProductEndpoint(req: Request, res: Response): Promise<void> {
    const appendObject = (object1: object, object2: object) => {
        return {...object1, ...object2};
    };

    const postData = req.body;
    let updateData = {};
    let failedValidation: boolean = false;

    // One downside to this is it has hardcoded the values away from the actual class. Maybe revise this later.
    const keys: Record<string, any> = {
        "identifier": Product.stringIsUrlSafe,
        "name": Product.stringIsValid,
        "price": Product.numberIsValid,
        "category": Product.stringIsValid,
        "sizes": Product.arrayIsValid
    };

    Object.keys(postData).forEach((key) => {
        if (keys.hasOwnProperty(key)) {
            // Execute validation function tied to the key in the above object
            const value = postData[key];
            const validationFunction = (keys)[key];
            if (validationFunction(value)) {
                const newData: Record<string, any> = {};
                newData[key] = value;
                updateData = appendObject(updateData, newData);
            } else {
                // One of the fields could not be validated
                failedValidation = true;
            }
        }
    });

    if (failedValidation){
        return res.status(400).end("Could not update product as non-valid data was provided.");
    }

    if (updateData) {
        const updatedProduct = await ProductModel.updateOne({identifier: req.params.identifier}, updateData);
        if (updatedProduct) {
            res.status(200).end(`${req.params.identifier} was successfully updated.`);
        } else {
            // No product was found in the database, nothing was updated
            res.status(404).end(`${req.params.identifier} could not be found.`);
        }

    } else {
        return res.status(400).end("Could not update product as no valid fields were given.");
    }
}

/**
 * Deletes the product with the identifier given.
 *
 * ~~Responses~~
 * `204 No Content` on delete success.
 * `401 Unauthorized` on request without header `X-Token` being set to an authorized key. See checkAuthorization middleware function.
 * `404 Not Found` if product with identifier doesn't exist.
 * @param req - request object
 * @param res - response object
 */
async function deleteProductEndpoint(req: Request, res: Response): Promise<void> {
    const product = await ProductModel.findOneAndDelete({identifier: req.params.identifier});
    if (!product) {
        // If something was found and deleted, it returns the deleted object
        // If no return, then the object wasn't in the database!
        return res.status(404).end(`Product with identifier ${req.params.identifier} could not be found.`);
    } else {
        return res.status(204).end(`${req.params.identifier} has been deleted.`);
    }
}

/**
 * Defines all endpoints with middleware
 * @param server - express Application to register endpoints on.
 */
function defineEndpoints(server: Application): void {
    // REGISTER API ROUTES
    const apiBaseEndpoint = "/api/products";
    const specifiedProductEndpoint = `${apiBaseEndpoint}/:identifier`;
    // Middleware to parse json in raw body requests
    const jsonParser = jsonBodyParser();

    // GET all products
    server.get(apiBaseEndpoint, getProductsEndpoint);

    // GET specified product
    server.get(specifiedProductEndpoint, getProductEndpoint);

    // POST create new product
    // middleware - Check if client is authorized, Parse JSON from body, then try and create product
    server.post(apiBaseEndpoint, [helpers.checkAuthorization, jsonParser, createProductEndpoint]);

    // PATCH edit specified product
    // middleware - Check if client is authorized, Parse JSON from body, then try and edit product
    server.patch(specifiedProductEndpoint, [helpers.checkAuthorization, jsonParser, editProductEndpoint]);

    // DELETE specified product
    // middleware - Check if client is authorized, then delete product
    server.delete(specifiedProductEndpoint, [helpers.checkAuthorization, deleteProductEndpoint]);
}

/**
 * Runs REST API server. Makes sure the connection to the mongodb is established.
 */
export function runServer(): void {
    connectDB();
    const server = express();
    server.use(morgan("dev"));
    server.use(helpers.handleMiddlewareError);

    defineEndpoints(server);
    // Run Server
    server.listen(8081, () => {
        console.log("API running @ http://127.0.0.1:8081");
    });
}
