
import { arrayProp, prop, getModelForClass } from '@typegoose/typegoose';

/**
 * @class Base class for models with helper functions to validate JSON inputs
 */
class BaseModel {
    /**
     * Validates sizes array, making sure it either is undefined, or defined. All items have to be strings.
     * @param possibleArray - Object to run tests against. Can be anything.
     * @returns true or false to whether the object is a valid array
     */
    public static arrayIsValid(possibleArray: any): boolean {
        if (Array.isArray(possibleArray)) {
            const allItemsStrings = possibleArray.every( (value) => { return typeof value === "string"; });
            if (!allItemsStrings) {
                return false;
            } else {
                return true;
            }
        } else if (typeof possibleArray === "undefined") {
            // No array provided is valid
            return true;
        } else {
            // sizes is not array but also not undefined
            return false;
        }
    }

    /**
     * Validates strings. Checks if they are a string and that they are not empty.
     * @param possibleArray - Object to run tests against. Can be anything.
     * @returns true or false to whether the object is a valid string
     */
    public static stringIsValid(possibleString: any): boolean {
        if (typeof possibleString === "string") {
            // Aka if not empty
            if (possibleString.replace(" ", "")) {
                return true;
            }
        }
        return false;
    }

    /**
     * Takes a string and checks if it only contains alphanumeric characters
     * @param str - string to check if url safe. Could be anything, hence the check during the function.
     * @returns boolean of if string is alphanumeric. If not a string, returns false.
     */
    public static stringIsUrlSafe(str: any): boolean {
        const containsIllegalChars = new RegExp(/[^a-z0-9\-\_]/gi);
        if (containsIllegalChars.test(str)) {
            return false;
        }
        return true;
    }

    /**
     * Checks if number is a number object and is positive.
     * @param str - number to check. Could be anything
     * @returns boolean of if number is a positive integer
     */
    public static numberIsValid(int: any): boolean {
        if (typeof int === "number") {
            if (int >= 0) { return true; }
        }
        // For simplicity, we will reject strings that can be converted to numbers for now
        return false;
    }
}


/**
 * @class A interface for the Product schema. The model is derived from the properties and their decorators.
 * The class inherits helper methods to validate itself and its properties.
 */
export class Product extends BaseModel {
    @prop({ required: true, unique: true, dropDupes: true })
    public identifier!: string;
    @prop({ required: true })
    public name!: string;
    @prop({ required: true })
    public price!: number;
    @arrayProp({ items: String })
    public sizes?: string[];
    @prop({ required: true })
    public category!: string;

    /**
     * Validates object if it can be cast to product interface.
     * Checks if all required fields are provided and that the data meets the guidelines.
     * If extra fields given, we will just ignore them. Possible issue here because we are failing quietly.
     * @param jsonInput - Object to be casted to Product interface
     * @throws error if a received part of data is not valid and cannot be parsed
     * @returns Product interface with valid fields and data
     */
    public static fromJson(jsonInput: any): Product {
        // Check if keys exist and types are correct
        const isValid =
            this.stringIsUrlSafe(jsonInput.identifier) &&
            this.stringIsValid(jsonInput.name) &&
            this.numberIsValid(jsonInput.price) &&
            this.stringIsValid(jsonInput.category);

        const arrayIsValid = this.arrayIsValid(jsonInput.sizes);

        if (!isValid || !arrayIsValid) {
            throw new Error("POST data is missing required keys or data provided is not valid to the schema.");
        }

        // Build object to return
        const validatedProduct: Product = {
            identifier: jsonInput.identifier,
            name: jsonInput.name,
            price: jsonInput.price,
            category: jsonInput.category,
        };

        if (arrayIsValid) {
            validatedProduct.sizes = jsonInput.sizes;
        }

        return validatedProduct;
    }
}

export const ProductModel = getModelForClass(Product);
