import { Application, NextFunction, Request, Response } from "express";


/**
 * Checks if an error has occurred and if so, return 500 to the user and don't continue processing
 * @param err - error object
 * @param req - request object
 * @param res - response object
 * @param next - next function in the chain
 */
export function handleMiddlewareError(err: Error, req: Request, res: Response, next: NextFunction) {
    if (err) {
        // General catch-all for any error in the middleware. Usually this is failing to parse provide JSON at all.
        res.status(500).end("Something went wrong!");
        return;
    }
    // If no error, just continue
    next();
}

/**
 * Middleware for the server to have a simple authentication system.
 * Checks headers to see if "X-Token" is set to "SECRET_KEY". If it isn't, then the server will return a 401 due to lack of authentication.
 * @param req - request object
 * @param res - response object
 * @param next - next function in the chain
 */
export function checkAuthorization(req: Request, res: Response, next: NextFunction) {
    const token: string | undefined = req.get("X-Token");
    if (token === "SECRET_KEY") {
        next();
    } else {
        res.status(401).end(); // Return 401 Unauthorized due to no authentication
    }
}

/**
 * Converts the integer parameters in the API requests
 * @param param - the string parameter passed by the user
 * @throws Error if the string cannot be converted into a positive integer and thus can't be used to filter the database query.
 */
export function parseParamInt(param: string): number {
    const int = parseInt(param, 10);
    if ( isNaN(int) || int < 0 ) {
        throw new Error("Param cannot be processed into a valid number");
    }
    return int;
}
